# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A Config that builds a `dict` and accepts any parameters."""

from fiddle import config


def _kwargs_to_dict(**kwargs):
  return dict(kwargs)


class DictConfig(config.Config):
  """A Config instance that builds a `dict` accepting any keys/values."""

  def __init__(self, *args, **kwargs):
    super().__init__(_kwargs_to_dict, *args, **kwargs)
