# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Init file for the `fiddle` package."""

from fiddle.building import build
from fiddle.building import BuildError
from fiddle.config import assign
from fiddle.config import Buildable
from fiddle.config import Config
from fiddle.config import Partial
from fiddle.config import update_callable
from fiddle.materialize import materialize_defaults
from fiddle.tagging import set_tagged
from fiddle.tagging import Tag
from fiddle.tagging import TaggedValue
