# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Types to make up some very important business logic.

These are types that will be configured via Fiddle's command line args.

Importantly: there is no dependency on Fiddle in this file!
"""

import dataclasses


@dataclasses.dataclass
class MyLinearModel:
  w: float
  b: float
  activation_dtype: str

  def __call__(self, x: float) -> float:
    return x * self.w + self.b


class MyDataLoader:

  def __init__(self, filename: str):
    pass

  def __iter__(self):
    return iter([1.0])  # Empty iterator


@dataclasses.dataclass
class BulkInferenceRunner:
  model: MyLinearModel
  data: MyDataLoader

  def run(self):
    print('Running...')
    for example in self.data:
      print(self.model(example))
    print('All done!')
